# -*- coding: utf-8 -*-
# __main__.py
# Authors: Stephan Meighen-Berger

def main():
    """ Tells the user how to use fennel
    """
    print("Welcome to fennel!")
    print("Please note fennel is a package and cannot be directly executed")
    print("The current fennel version is 1.2.2")
    print("Usage after installation:")
    print("from fennel import Fennel")
    print("fennel_obj = Fennel()")
    print("Visit the following link for examples:")
    print("https://github.com/MeighenBergerS/fennel/tree/master/notebooks")


if __name__ == "__main__":
    main()
