# -*- coding: utf-8 -*-

from .fennel import Fennel
from .config import config

__all__ = (Fennel, config)

# Version of the fennel package
__version__ = "1.2.2"
__author__ = "Stephan Meighen-Berger"
